# frozen_string_literal: true

# The Agent interacts with the Web Application and Profiler.
# It is also responsible in requesting the instrumenter to build the transactions and spans.

require 'stackify_apm/naively_hashable'
require 'stackify_apm/context_builder'
require 'stackify_apm/error_builder'
require 'stackify_apm/stacktrace_builder'
require 'stackify_apm/error'
require 'stackify_apm/trace_logger'
require 'stackify_apm/spies'
require 'stackify_apm/serializers'
require 'stackify_apm/worker'
require 'stackify_apm/util/trace_log_watcher'

module StackifyRubyAPM
  # @api private
  class Agent
    include Log

    LOCK = Mutex.new

    def self.instance # rubocop:disable Style/TrivialAccessors
      @instance
    end

    # rubocop:disable Metrics/PerceivedComplexity
    # rubocop:disable Metrics/CyclomaticComplexity
    def self.start(config)
      return @instance if @instance

      config = Config.new(config) unless config.is_a?(Config)
      case config.transport.downcase
      # For unix socket we don't create a trace log file
      when StackifyRubyAPM::TRACE_LOG
        # For Default transport let's create initial trace log file
        pid = $PID || Process.pid
        host_name = (config.hostname || `hostname`).strip
        date_now = Time.now
        current_trace_file = config.log_trace_path + squish(host_name) + '#' + squish(pid.to_s)
        if ENV['STACKIFY_RUBY_ENV'] != 'rspec'
          config.tracer_logger = StackifyLogger.new(current_trace_file, config.filenum_rotate, config.logger_byte_size)
          config.debug_logger
          config.logtime_created = date_now.strftime('%H:%M')
          Util::TraceLogWatcher.delete_trace_logs(config)
        end
      end

      LOCK.synchronize do
        return @instance if @instance

        @instance = new(config).start
      end
    end
    # rubocop:enable Metrics/CyclomaticComplexity
    # rubocop:enable Metrics/PerceivedComplexity

    def self.stop
      LOCK.synchronize do
        return unless @instance

        @instance.stop
        @instance = nil
      end
    end

    def stop
      @instrumenter.stop
      kill_worker
      self
    end

    def self.running?
      !@instance.nil?
    end

    def initialize(config)
      @config = config
      @trace_logger = TraceLogger.new(config)
      @messages = Queue.new
      @pending_transactions = Queue.new
      @instrumenter = Instrumenter.new(self)
      @context_builder = ContextBuilder.new(self)
      @error_builder = ErrorBuilder.new(self)
      @stacktrace_builder = StacktraceBuilder.new(self)
      @error_serializer = Serializers::Errors.new(config)
    end

    attr_reader :config,
                :messages,
                :pending_transactions,
                :instrumenter,
                :context_builder,
                :stacktrace_builder,
                :trace_logger,
                :error_serializer,
                :error_builder

    def start
      info '[Agent] start()'
      info '[Agent] transport type: ' + @config.transport
      spies_name = ''
      return false unless @config.instrument
      config.enabled_spies.each do |lib|
        spies_name = spies_name + ', ' + lib.inspect.to_s

        require "stackify_apm/spies/#{lib}"
      end

      debug '[Agent] Loaded spies: ' + spies_name if ENV['STACKIFY_TRANSPORT_LOG_LEVEL'] == '0'
      self
    end

    # queues
    # Stores transaction in queue
    #
    def enqueue_transaction(transaction)
      if !@config.queue
        return @trace_logger.post([transaction])
      end

      boot_worker unless worker_running?
      pending_transactions.push(transaction)
      return unless should_flush_transactions?

      messages.push(Worker::FlushMsg.new)
    end

    def should_flush_transactions?
      return true unless config.flush_interval_seconds
      return true if pending_transactions.length >= config.max_queue_size

      false
    end

    def enqueue_error(error)
      boot_worker unless worker_running?

      messages.push(Worker::ErrorMsg.new(error))
    end

    # Instrumentation
    #
    def current_transaction
      instrumenter.current_transaction
    end

    # Loads transaction
    #
    def transaction(*args, &block)
      instrumenter.transaction(*args, &block)
    end

    def span(*args, &block)
      instrumenter.span(*args, &block)
    end

    # Responsible for building the transaction's context
    #
    def build_context(rack_env)
      @context_builder.build(rack_env)
    end

    # errors

    def report(exception, handled: true)
      return if config.filter_exception_types.include?(exception.class.to_s)

      error = @error_builder.build_exception(
        exception,
        handled: handled
      )
      current_error = @error_serializer.build(error)
      current_transaction.add_exception(current_error) if current_transaction
    end

    def report_message(message, backtrace: nil, **attrs)
      error = @error_builder.build_log(
        message,
        backtrace: backtrace,
        **attrs
      )
      enqueue_error error
    end

    # This method will strip and remove garbage character and multiple spaces.

    def self.squish(str)
      str = str.sub(/\A[[:space:]]+/, '')
      str = str.sub(/[[:space:]]+\z/, '')
      str = str.sub(/[[:space:]]+/, ' ')
      str = str.strip.sub(/\s+/, '')
      str
    end

    private

    def boot_worker
      debug '[Agent] Booting worker' if ENV['STACKIFY_TRANSPORT_LOG_LEVEL'] == '0'

      @worker_thread = Thread.new do
        Worker.new(
          config,
          messages,
          pending_transactions,
          trace_logger
        ).run_forever
      end
    end

    def kill_worker
      messages << Worker::StopMsg.new
      raise 'Failed to wait for worker, not all messages sent' if @worker_thread && !@worker_thread.join(5) # 5 secs

      @worker_thread = nil
    end

    def worker_running?
      @worker_thread && @worker_thread.alive?
    end
  end
end
