# frozen_string_literal: true

#
# The Config class sets the APM's default values.
#

require 'logger'
require 'yaml'
require 'socket'
module StackifyRubyAPM
  TRANSPORT = [
    TRACE_LOG = 'default'.freeze,
    UNIX_SOCKET = 'agent_socket'.freeze,
    AGENT_HTTP = 'agent_http'.freeze,
    LOGGING = 'logging'.freeze
].freeze

  # @api private
  class Config
    include Log
    DEFAULTS = {
      config_file: 'config/stackify_apm.yml',
      json_config_file: 'config/stackify.json',
      rum_script_src: 'https://stckjs.azureedge.net/stckjs.js',
      application_name: 'Ruby Application',
      environment_name: ENV['RAILS_ENV'] || ENV['RACK_ENV'],
      already_instrumented_flag: false,
      rum_auto_injection: false,
      rum_enabled: true,
      rum_cookie_path: '/',
      rum_cookie_name: '.Stackify.Rum',
      transport: StackifyRubyAPM::TRACE_LOG,
      instrument: true,
      debug_logging: false,
      log_path: StackifyRubyAPM::Util.host_os == 'WINDOWS' ? 'C:\ProgramData\Stackify\Ruby\log\stackify-ruby-apm-1.log' : '/usr/local/stackify/stackify-ruby-apm/log/stackify-ruby-apm-1.log',
      log_level: Logger::INFO,
      log_trace_path: StackifyRubyAPM::Util.host_os == 'WINDOWS' ? 'C:\ProgramData\Stackify\Ruby\log\\' : '/usr/local/stackify/stackify-ruby-apm/log/',

      max_queue_size: 500, # Maximum queue length of transactions before sending transactions to the APM.
      flush_interval_seconds: 1, # interval with which transactions should be sent to the APM. Default value: 1 seconds
      delay_seconds: 10, # Interval in seconds between tries
      max_retries: 3, # Number of attempts

      filter_exception_types: [],

      tracer_logger: nil,
      logger_byte_size: 50_000_000, # e.g, 50_000_000=50mb
      filenum_rotate: 20,
      debugger_byte_size: 20_000_000, # e.g, 20_000_000=20mb
      debugger_filenum_rotate: 3,
      logtime_created: 0,
      http_status: nil,
      debug_transactions: false,
      trace_log_age_in_hour: 60 * 60, # e.g, 60*60=1hour, 60*2=2mins
      check_trace_log_per_hour: '1h', # e.g, 1h=every 1hour, 2m=every 2mins

      source_lines_error_app_frames: 5,
      source_lines_span_app_frames: 5,
      source_lines_error_library_frames: 0,
      source_lines_span_library_frames: 0,
      span_frames_min_duration: -1, # it will collect stack traces for all spans

      disabled_spies: %w[json],

      view_paths: [],
      root_path: Dir.pwd,
      stackify_properties_file: '/usr/local/stackify/stackify-ruby-apm/stackify.properties',
      agent_traces_url: '/traces',
      unix_socket_path: '/usr/local/stackify/stackify.sock',
      transport_http_endpoint: 'https://localhost:10601',

      queue: true,
      lambda_handler: '',
      prefix_enabled: false
    }.freeze

    ENV_TO_KEY = {
      'STACKIFY_DEBUG' => [:bool, 'debug_logging'],
      'STACKIFY_APPLICATION_NAME' => 'application_name',
      'STACKIFY_ENVIRONMENT_NAME' => 'environment_name',
      'STACKIFY_RUM' => [:bool, 'rum_enabled'],
      'STACKIFY_RUM_AUTO_INJECT' => [:bool, 'rum_auto_injection'],
      'STACKIFY_RUM_SCRIPT_SRC' => 'rum_script_src',
      'STACKIFY_TRANSPORT' => 'transport',
      'STACKIFY_TRANSPORT_LOG_PATH' => 'log_path',
      'STACKIFY_TRANSPORT_LOG_LEVEL' => [:int, 'log_level'],
      'STACKIFY_TRANSPORT_SOCKET_PATH' => 'unix_socket_path',
      'STACKIFY_TRANSPORT_HTTP_ENDPOINT' => 'transport_http_endpoint',
      'STACKIFY_ALREADY_INSTRUMENTED_FLAG' => 'already_instrumented_flag',
      'STACKIFY_INSTRUMENT' => [:bool, 'instrument'],
      'STACKIFY_HOSTNAME' => 'hostname',
      'STACKIFY_SOURCE_LINES_ERROR_APP_FRAMES' => [:int, 'source_lines_error_app_frames'],
      'STACKIFY_SOURCE_LINES_SPAN_APP_FRAMES' => [:int, 'source_lines_span_app_frames'],
      'STACKIFY_SOURCE_LINES_ERROR_LIBRARY_FRAMES' => [:int, 'source_lines_error_library_frames'],
      'STACKIFY_SOURCE_LINES_SPAN_LIBRARY_FRAMES' => [:int, 'source_lines_span_library_frames'],
      'STACKIFY_SPAN_FRAMES_MIN_DURATION' => [:int, 'span_frames_min_duration'],
      'STACKIFY_MAX_QUEUE_SIZE' => [:int, 'max_queue_size'],
      'STACKIFY_FLUSH_INTERVAL' => 'flush_interval_seconds',
      'STACKIFY_DISABLED_SPIES' => [:list, 'disabled_spies'],
      'STACKIFY_QUEUE' => [:bool, 'queue'],
      'STACKIFY_LAMBDA_HANDLER' => 'lambda_handler',
      'STACKIFY_PREFIX_ENABLED' => [:bool, 'prefix_enabled']
    }.freeze

    def initialize(options = {})
      set_defaults
      set_from_args(options)
      set_from_config_file
      set_from_env
      set_prefix_paths if @prefix_enabled
      yield self if block_given?
      debug_logger
      StackifyRubyAPM::Util.host_os == 'WINDOWS' ? load_stackify_props_windows : load_stackify_props_linux
      validate_apm_yml
    end

    attr_accessor :config_file
    attr_accessor :json_config_file
    attr_accessor :environment_name
    attr_accessor :rum_script_src
    attr_accessor :rum_auto_injection
    attr_accessor :rum_enabled
    attr_accessor :rum_cookie_path
    attr_accessor :rum_cookie_name
    attr_accessor :already_instrumented_flag
    attr_accessor :rum_script_injected
    attr_accessor :instrument
    attr_accessor :enabled_environments
    attr_accessor :stackify_properties_file
    attr_accessor :transport
    attr_accessor :agent_traces_url
    attr_accessor :unix_socket_path
    attr_accessor :transport_http_endpoint

    attr_accessor :application_name
    attr_accessor :hostname

    attr_accessor :log_path
    attr_accessor :log_level
    attr_accessor :logger
    attr_accessor :debug_logging
    attr_accessor :logger_byte_size
    attr_accessor :filenum_rotate
    attr_accessor :debugger_byte_size
    attr_accessor :debugger_filenum_rotate
    attr_accessor :trace_log_age_in_hour
    attr_accessor :check_trace_log_per_hour

    attr_accessor :tracer_logger
    attr_accessor :logtime_created

    attr_accessor :log_trace_path
    attr_accessor :source_lines_error_app_frames
    attr_accessor :source_lines_span_app_frames
    attr_accessor :source_lines_error_library_frames
    attr_accessor :source_lines_span_library_frames
    attr_accessor :span_frames_min_duration

    attr_accessor :max_queue_size
    attr_accessor :flush_interval_seconds
    attr_accessor :delay_seconds
    attr_accessor :max_retries

    attr_accessor :filter_exception_types

    attr_accessor :debug_transactions

    attr_accessor :disabled_spies

    attr_accessor :view_paths
    attr_accessor :root_path
    attr_accessor :http_status

    attr_accessor :queue
    attr_accessor :lambda_handler
    attr_accessor :prefix_enabled

    attr_reader :client_id
    attr_reader :device_id
    attr_reader :apm_disabled_in_rake
    attr_reader :client_run_domain

    def app=(app)
      case app_type?(app)
      when :rails
        set_rails(app)
      else
        # TODO: define custom?
        self.application_name = 'ruby'
      end
    end

    def app_type?(app)
      return :rails if defined?(::Rails) && app.is_a?(Rails::Application)

      nil
    end

    # available spies to use when framework is rails
    def available_spies
      %w[
        action_dispatch
        mongo
        net_http
        custom_instrumenter
        httpclient
        redis
        sequel
        tilt
        httprb
        curb
        curb/easy
        curb/multi
        httparty
        stackify_logger
        sidekiq
        delayed_job 
        faraday
        sucker_punch
        dynamo_db
      ]
    end

    def prefix_spies
      return [] unless @prefix_enabled
      %w[
        logger
        logging
        log4r
        yell
      ]
    end

    def enabled_spies
      # check if the framework is rails or sinatra
      sinatra_activerecord_spies = %w[
        sinatra
        sinatra_activerecord/mysql_adapter
        sinatra_activerecord/postgresql_adapter
        sinatra_activerecord/sqlite_adapter
      ]

      new_available_spies = if defined? ::Sinatra::Base
                              if defined? ActiveRecord
                                available_spies + sinatra_activerecord_spies
                              else
                                available_spies
                              end
                            else
                              available_spies
                            end

      new_available_spies + prefix_spies - disabled_spies
    end

    # Default Transport
    #
    # initialize default logger transport
    def debug_logger
      case @transport.downcase
      # For unix socket we don't create a trace log file
      when StackifyRubyAPM::TRACE_LOG
        debugger_logpath = log_path == '-' ? $stdout : log_path
        logger = StackifyLogger.new(debugger_logpath, debugger_filenum_rotate, debugger_byte_size)
        logger.level = log_level
        self.logger = logger
      end
    end

    private

    def assign(options)
      options.each do |key, value|
        begin
          send("#{key}=", value)
        rescue Exception => e
          info "[Config] Key: '#{key}' doesn't exist."
        end
      end
    end

    def set_defaults
      assign(DEFAULTS)
    end

    # rubocop:disable Metrics/CyclomaticComplexity
    def set_from_env
      ENV_TO_KEY.each do |env_key, key|
        next unless (value = ENV[env_key])

        type, key = key if key.is_a? Array

        value =
          case type
          when :int then value.to_i
          when :float then value.to_f
          when :bool then !%w[0 false].include?(value.strip.downcase)
          when :list then value.split(/[ ,]/)
          else value
          end

        send("#{key}=", value)
      end
    end
    # rubocop:enable Metrics/CyclomaticComplexity

    # rubocop:disable Naming/AccessorMethodName
    def set_from_args(options)
      assign(options)
    end
    # rubocop:enable Naming/AccessorMethodName

    def set_from_config_file
      return unless File.exist?(config_file)

      assign(YAML.load_file(config_file) || {})
    end

    # set log paths to prefix path if prefix_enabled is true
    def set_prefix_paths
      all_user_profile = ENV['ALLUSERSPROFILE'] || "C:\\ProgramData"

      @log_path = StackifyRubyAPM::Util.host_os == 'WINDOWS' ? "#{all_user_profile}\\Stackify\\Agent\\debug\\stackify-ruby-apm-1.log" : "/usr/local/prefix/debug/stackify-ruby-apm-1.log"
      @log_trace_path = StackifyRubyAPM::Util.host_os == 'WINDOWS' ? "#{all_user_profile}\\Stackify\\Agent\\log\\" : '/usr/local/prefix/log/'
    end

    # rubocop:disable Naming/AccessorMethodName
    def set_rails(app)
      self.application_name ||= format_name(application_name || app.class.parent_name).strip
      self.logger ||= Rails.logger

      self.root_path = Rails.root.to_s
      self.view_paths = app.config.paths['app/views'].existent
    end
    # rubocop:enable Naming/AccessorMethodName

    def build_logger
      debugger_logpath = log_path == '-' ? $stdout : log_path
      logger = StackifyLogger.new(debugger_logpath, debugger_filenum_rotate, debugger_byte_size)
      logger.level = log_level
      self.logger = logger
    end

    def format_name(str)
      str.gsub('::', '_')
    end

    # rubocop:disable Metrics/CyclomaticComplexity
    def load_stackify_props_linux
      @client_id = nil
      @device_id = nil
      @client_run_domain = nil
      begin
        info '[Config] Reading the stackify.properties file'
        IO.foreach(@stackify_properties_file) do |line|
          case line.downcase
          when /clientid=\d+/
            cid = line.split('=')[1].strip
            @client_id = "C#{cid}"
          when /deviceid=\d+/
            devid = line.split('=')[1].strip
            @device_id = "CD#{devid}"
          when /clientrumdomain=([\w\s\d\"\']+)+/i
            @client_run_domain = line.split('=')[1].strip
          end
        end
        info "[Config] stackify.properties: clientId=#{@client_id}, deviceId=#{@device_id}, client_rum_domain=#{@client_run_domain}"
      rescue StandardError => e
        info '[Config] Error occured while reading the stackify.properties file.'
        info e.inspect
      end
      info '[Config] No clientId found from stackify.properties file.' if @client_id.nil?
      info '[Config] No deviceId found from stackify.properties file.' if @device_id.nil?
      info '[Config] No client RUM domain found from stackify.properties file.' if @client_run_domain.nil?
    end
    # rubocop:enable Metrics/CyclomaticComplexity

    # rubocop:disable Metrics/CyclomaticComplexity
    # rubocop:disable Metrics/PerceivedComplexity
    def load_stackify_props_windows
      require 'win32ole'
      @client_id = nil
      @device_id = nil
      @client_run_domain = nil
      begin
        info '[Config] Reading the Stackify props in Environment variables'
        @client_run_domain = ENV['STACKIFY_RUM_DOMAIN'] || nil
        if ENV['STACKIFY_ENV']
          props = ENV['STACKIFY_ENV'].split('|')
          @client_id = props[0].to_s if props[0]
          @device_id = props[1].to_s if props[1]
          info "[Config] stackify.properties: clientId=#{@client_id}, deviceId=#{@device_id}, client_rum_domain=#{@client_run_domain}"
        end
      rescue StandardError => e
        info '[Config] Error occured while reading the Stackify props in Environment variables.'
        info e.inspect
      end
      info '[Config] No clientId found in environment variables.' if @client_id.nil?
      info '[Config] No deviceId found in environment variables.' if @device_id.nil?
      info '[Config] No client RUM domain found in environment variables.' if @client_run_domain.nil?
    end
    # rubocop:enable Metrics/CyclomaticComplexity
    # rubocop:enable Metrics/PerceivedComplexity

    # rubocop:disable Metrics/CyclomaticComplexity
    # rubocop:disable Metrics/PerceivedComplexity
    def validate_apm_yml
      info '[Config] rum_enabled must be Boolean type: true/false.' unless [TrueClass, FalseClass].include?(@rum_enabled.class) && defined?(@rum_enabled)
      info '[Config] rum_auto_injection must be Boolean type: true/false.' unless [TrueClass, FalseClass].include?(@rum_auto_injection.class) && defined?(@rum_auto_injection)
      info '[Config] application_name must be String type.' unless @application_name.is_a?(String) && defined?(@application_name)
      info '[Config] environment_name must be String type.' unless @environment_name.is_a?(String) && defined?(@environment_name)
      info '[Config] transport must be String type.' unless @transport.is_a?(String) && defined?(@transport)
      info '[Config] Transport should be one of these values: [agent_socket, default, agent_http]. Should be a String.' if defined?(@transport) && !%w[agent_socket default agent_http].include?(@transport.downcase)
      info '[Config] prefix_enabled must be Boolean type: true/false.' unless [TrueClass, FalseClass].include?(@prefix_enabled.class) && defined?(@prefix_enabled)
    end
    # rubocop:enable Metrics/CyclomaticComplexity
    # rubocop:enable Metrics/PerceivedComplexity
  end
end
