# frozen_string_literal: true

require 'stackify_apm/span'
require 'stackify_apm/transaction'
#
# The Instrumenter will build/process the transactions and spans.
#
module StackifyRubyAPM
  # @api private
  class Instrumenter
    include Log

    KEY = :_stackify_transaction_key
    # Transaction Info class
    class TransactionInfo
      def initialize
        self.current = nil
      end

      def current
        Thread.current[KEY]
      end

      def current=(transaction)
        Thread.current[KEY] = transaction
      end
    end

    def initialize(agent)
      debug '[Instrumenter] initialize()' if ENV['STACKIFY_TRANSPORT_LOG_LEVEL'] == '0'
      @agent = agent
      @config = agent.config

      @transaction_info = TransactionInfo.new
    end

    attr_reader :agent, :config, :pending_transactions

    def stop
      current_transaction.release if current_transaction
      @subscriber.unregister! if @subscriber
    end

    def subscriber=(subscriber)
      @subscriber = subscriber
      @subscriber.register!
    end

    def current_transaction
      @transaction_info.current
    end

    def current_transaction=(transaction)
      @transaction_info.current = transaction
    end

    # Creates a new transaction or return the currently running
    #
    def transaction(*args)
      unless config.instrument
        yield if block_given?
        return
      end

      if (transaction = current_transaction)
        yield transaction if block_given?
        return transaction
      end

      transaction = Transaction.new self, *args

      self.current_transaction = transaction
      return transaction unless block_given?

      begin
        yield transaction
      ensure
        self.current_transaction = nil
        transaction.done
      end

      transaction
    end

    def span(*args, &block)
      unless current_transaction
        return yield if block_given?

        return
      end

      current_transaction.span(*args, &block)
    end

    # Once the transaction is submitted it will be stored temporarily in queue
    #
    def submit_transaction(transaction)
      debug '[Instrumenter] submit_transaction(transaction) transaction: ' + transaction.inspect.to_s if ENV['STACKIFY_TRANSPORT_LOG_LEVEL'] == '0'
      agent.enqueue_transaction transaction
      return unless config.debug_transactions
    end

    def inspect
      '<StackifyRubyAPM::Instrumenter ' \
        "current_transaction=#{current_transaction.inspect}" \
        '>'
    end
  end
end
