# frozen_string_literal: true

#
# This class monkey patches the Rack Middleware method: e.g., call(env)
# It also handles the submission of transactions to Agent.
#
# If Agent is running, build_transaction(env) is called:
#
#   -> BUILDS transaction: /stackify_ruby_apm.rb (#self.transaction)
#     -> builds a [Context] from a Rack `env`: /stackify_ruby_apm.rb (#self.build_context)
#     -> context may include information about the request,
#        response, current user and more
#
#   -> SUBMITS transaction: stackify_apm/transaction.rb (#submit)
#     -> instrumenter submits transaction
#     -> worker will run (if != running), responsible for sending transactions to APM
#        for every 1 secs (based on config.flush_interval_seconds)
#     -> transaction will be stored in queue by Agent
#     -> worker will constantly execute transaction sending to APM
#
require 'stackify_apm/response_manipulator'

module StackifyRubyAPM
  # @api private
  class Middleware
    attr_accessor :rack_body, :rack_headers, :rack_status, :configuration

    CONTENT_TYPE_REGEX = %r{text\/html|application\/xhtml\+xml/}
    CONTENT_DISPOSITION = 'Content-Disposition'.freeze
    ATTACHMENT = 'attachment'.freeze

    def initialize(app)
      @app = app
    end

    # This is where the requests are received, built into a transaction, and then submitted
    #
    # rubocop:disable Metrics/CyclomaticComplexity
    # rubocop:disable Metrics/PerceivedComplexity
    def call(env)
      begin
        context = StackifyRubyAPM.build_context(env)
        transaction = build_transaction(env, context) if running?
        resp = @app.call env

        @rack_status = resp[0].to_i
        @rack_headers = resp[1]
        @rack_body = resp[2]
        @configuration = config

        build_prefix_context(transaction, context)

        if okay_to_modify?
          @configuration.already_instrumented_flag = true
          if @configuration.rum_enabled.is_a?(TrueClass)
            # close old response body proxy to close all db connections
            @rack_body.close if @rack_body.respond_to?(:close)
            response = Rack::Response.new @rack_body, @rack_status, @rack_headers
            response.set_cookie(@configuration.rum_cookie_name, value: transaction.id, path: @configuration.rum_cookie_path)
            resp = response.finish
          end
          if @configuration.rum_auto_injection.is_a?(TrueClass)
            response_manupulate = StackifyRubyAPM::ResponseManipulator.new(env, resp, @configuration)
            response_string = response_manupulate.adjust_pagehtml_response
            if response_string
              response = Rack::Response.new(response_string, resp[0].to_i, resp[1])
              resp = response.finish
            else
              resp
            end
          else
            resp
          end
        else
          resp
        end

        submit_transaction(transaction, *resp) if transaction
      rescue InternalError
        raise # Don't report StackifyRubyAPM errors
      rescue StandardError
        transaction.submit('500', status: 500) if transaction
        raise
      ensure
        transaction.release if transaction
      end

      resp
    end
    # rubocop:enable Metrics/PerceivedComplexity
    # rubocop:enable Metrics/CyclomaticComplexity

    def submit_transaction(transaction, status, headers, _body)
      result = status.to_i
      transaction.submit(result, status: status, headers: headers)
    end

    # Start of transaction building with params: name, type, context
    #
    def build_transaction(env, context=nil)
      context = context || StackifyRubyAPM.build_context(env)
      StackifyRubyAPM.transaction 'Rack', 'WEBAPP', context: context
    end

    def running?
      StackifyRubyAPM.running?
    end

    def config
      StackifyRubyAPM.agent.config
    end

    def okay_to_modify?
      return false if xhr?
      return false unless modifiable_content_type?
      return false if @rack_status == 404 || @rack_status == 501
      return false if attachment?
      true
    end

    def modifiable_content_type?
      @rack_headers['Content-Type'] =~ self.class::CONTENT_TYPE_REGEX
    end

    def attachment?
      @rack_headers[CONTENT_DISPOSITION] && @rack_headers[CONTENT_DISPOSITION].include?(ATTACHMENT)
    end

    def streaming?(env)
      # Not used
      # Possible use of ActiveSupport.on_load(:action_controller_base)
      return false unless defined?(ActionController::Live)

      env['action_controller.instance'].class.included_modules.include?(ActionController::Live)
    end

    def xhr?
      @rack_headers['HTTP_X_REQUESTED_WITH'] == 'XMLHttpRequest'
    end

    # Add prefix instrument data to transaction context if enabled
    def build_prefix_context(transaction, context)
      return nil unless @configuration.prefix_enabled.is_a?(TrueClass)

      source = nil
      body = @rack_body
      body.each { |fragment| source ? (source << fragment.to_s) : (source = fragment.to_s) }

      transaction.context.prefix.request_body = context && context.request && context.request.body || ""
      transaction.context.prefix.request_headers = context && context.request && context.request.headers || Hash.new
      transaction.context.prefix.response_body = source || ""
      transaction.context.prefix.response_headers = @rack_headers || Hash.new
    end
  end
end
