# frozen_string_literal: true

#
# This class loads the specified config of the Rails app and starts the Agent.
#
# APM Config setting is done at stackify_apm/config.rb
#
# After the Agent has been started:
#   -> the enabled Spies (libraries) are registered
#   -> the Subscriber is called to register the events
#   -> the Middleware is then inserted
#

require 'stackify_apm/subscriber'

module StackifyRubyAPM
  # @api private
  # Railtie - core of the Rails framework and provides several hooks to modify the initialization process.
  class Railtie < Rails::Railtie
    config.stackify = ActiveSupport::OrderedOptions.new

    Config::DEFAULTS.each { |option, value| config.stackify[option] = value }

    # This will overwrite the default values of the config based from stackify_apm.yml initialized

    initializer 'stackify_apm.initialize' do |app|
      config = app.config.stackify.merge(app: app).tap do |c|
        # Prepend Rails.root to log_path if present
        c.log_path = Rails.root.join(c.log_path) if c.log_path && !c.log_path.start_with?('/')
      end

      begin
        agent = StackifyRubyAPM.start config
        if agent
          agent.instrumenter.subscriber = StackifyRubyAPM::Subscriber.new(agent)
          app.middleware.insert 0, Middleware # add a middleware to the application
        end
      rescue StandardError => e
        Rails.logger.error "[Railtie] Failed to start: #{e.message}"
      end
    end

    config.after_initialize do
      require 'stackify_apm/spies/action_dispatch'
    end
  end
end
