# frozen_string_literal: true

#
# This class will manupulate the html page d: e.g., call(env)
#
#

module StackifyRubyAPM
  # an abstraction for manipulating the HTML we capture in the middleware
  class ResponseManipulator
    include Log
    attr_reader :rack_response, :jsfile_to_inject
    attr_reader :rack_status, :rack_headers, :rack_body, :Rack_flagger
    attr_reader :env

    # examine in order to look for a RUM insertion point.
    SCAN_LIMIT = 50_000
    HEAD_START = '<head'.freeze
    HEAD_END = '</head'.freeze
    RUM_SCRIPT_VARIABLE = '[RUM_SCRIPT_VARIABLE'.freeze
    BRKT = ']'.freeze
    GT = '>'.freeze
    CHARSET_RE = /<\s*meta[^>]+charset\s*=[^>]*>/
    X_UA_COMPATIBLE_RE = /<\s*meta[^>]+http-equiv\s*=\s*['"]x-ua-compatible['"][^>]*>/

    def initialize(env, rack_response, config)
      @env = env
      @rack_response = rack_response

      @rack_status = rack_response[0]
      @rack_headers = rack_response[1]
      @rack_body = rack_response[2]
      @rack_flagger = nil
      @config = config
    end

    def rebuild_rack_response
      [rack_status, rack_headers, rack_body]
    end

    # rubocop:disable Metrics/CyclomaticComplexity
    # rubocop:disable Metrics/PerceivedComplexity
    def adjust_pagehtml_response
      response = @rack_body
      source = gather_source(response)
      close_old_response(response)
      return nil unless source

      client_id = @config.client_id
      device_id = @config.device_id
      client_rundomain = @config.client_run_domain
      inject_flag = false

      if client_id && device_id
        if StackifyRubyAPM.check_isdomain(client_rundomain)
          inject_flag = true
        else
          info 'RUM Injection Error: Client RUM Domain is invalid.'
        end
      else
        info 'RUM Injection Error: No Device ID and/or Client ID found.'
      end

      return unless inject_flag

      # Only scan the first 50k (roughly) then give up.
      insertion_index = find_end_of_head_open(source[0..SCAN_LIMIT])

      if insertion_index && inject_flag
        source = source[0...insertion_index] <<
                 StackifyRubyAPM.inject_rum_script <<
                 source[insertion_index..-1]
      end
      source
    end
    # rubocop:enable Metrics/CyclomaticComplexity
    # rubocop:enable Metrics/PerceivedComplexity

    def find_end_of_head_open(beginning_of_source)
      head_open = beginning_of_source.index(HEAD_END)
      beginning_of_source.index(GT, head_open) - 6 if head_open
    end

    def gather_source(response)
      source = nil
      response.each { |fragment| source ? (source << fragment.to_s) : (source = fragment.to_s) }
      source
    end

    def close_old_response(response)
      response.close if response.respond_to?(:close)
    end
  end
end
