# frozen_string_literal: true

#
# This module communicates and registers the additional or enabled outside library method
#

require 'forwardable'
require 'stackify_apm/util/inflector'

module StackifyRubyAPM
  # @api private
  module Spies
    # @api private
    class Registration
      extend Forwardable
      # rubocop:disable Metrics/ParameterLists
      def initialize(const_name, require_paths, spy, custom_instrumented = false, custom_method = '', tracked_func = false, tracked_function_name = '')
        @const_name = const_name
        @require_paths = Array(require_paths)
        @spy = spy
        @custom_instrumented = custom_instrumented
        @custom_method = custom_method
        @tracked_func = tracked_func
        @tracked_function_name = tracked_function_name
      end
      # rubocop:enable Metrics/ParameterLists

      attr_reader :const_name, :require_paths, :custom_instrumented, :custom_method, :tracked_func, :tracked_function_name

      def_delegator :@spy, :install
    end

    def self.require_hooks
      @require_hooks ||= {}
    end

    def self.installed
      @installed ||= {}
    end

    def self.register(*args)
      registration = Registration.new(*args)

      if safe_defined?(registration.const_name)
        registration.install
        installed[registration.const_name] = registration
      else
        register_require_hook registration
      end
    end

    def self.register_require_hook(registration)
      registration.require_paths.each do |path|
        require_hooks[path] = registration
      end
    end

    def self.hook_into(name)
      return unless (registration = require_hooks[name])
      return unless safe_defined?(registration.const_name)

      installed[registration.const_name] = registration
      registration.install

      registration.require_paths.each do |path|
        require_hooks.delete path
      end
    end

    def self.safe_defined?(const_name)
      Util::Inflector.safe_constantize(const_name)
    end

    def self.parse_json_config(file)
      jsondata = {}
      if ENV['STACKIFY_RUBY_ENV'] == 'rspec'
        file = 'spec/integration/stackify_spec.json'
      end
      return unless File.exist?(file)

      File.open(file) do |f|
        jsondata = JSON.parse(f.read)
      end
      jsondata
    end

    def self.class_exists?(class_name)
      if class_name
        klass = Module.const_get(class_name)
        return klass.is_a?(Class)
      end
    rescue NameError
      return false
    end

    # rubocop:disable Metrics/CyclomaticComplexity
    # rubocop:disable Metrics/PerceivedComplexity
    def self.ismethod_exists(class_name, current_method)
      module_consget = Module.const_get(class_name)
      current_method_without_apm = "_without_apm_#{current_method}"
      self_current_method_without_apm = "_self_without_apm_#{current_method}"

      with_flag = module_consget.instance_methods(false).include?(:"#{current_method_without_apm}")
      self_with_flag = module_consget.instance_methods(false).include?(:"#{self_current_method_without_apm}")

      current_method_exists = module_consget.instance_methods(false).include?(:"#{current_method}")
      self_current_method_exists = module_consget.instance_methods(false).include?(:"self.#{current_method}")

      exists = false
      if (self_with_flag == false && with_flag == false) &&
         (current_method_exists == true || self_current_method_exists == true)
        exists = true
      end

      if (module_consget.methods.include?(:"#{current_method_without_apm}") == false &&
         module_consget.methods.include?(:"#{self_current_method_without_apm}") == false) &&
         (module_consget.methods.include?(:"#{current_method}") || module_consget.methods.include?(:"self.#{current_method}"))
        exists = true
      end
      exists
    end
    # rubocop:enable Metrics/CyclomaticComplexity
    # rubocop:enable Metrics/PerceivedComplexity
  end
end

# @api private
module Kernel
  private

  alias require_without_apm require

  def require(path)
    res = require_without_apm(path)

    begin
      StackifyRubyAPM::Spies.hook_into(path)
    rescue StandardError => e
      puts "Failed hooking into '#{path}'. Please report this at " \
        'git@bitbucket.org:stackify/stackify-ruby-apm.git'
      puts e.backtrace.join("\n")
    end

    res
  end
end
