# frozen_string_literal: true

# Monkey patch for the custom instrumentation any values from config/stackify.json will be loop and will
# register as a spy.
#
require 'logger'
require 'stackify_apm/instrumenter_helper.rb'
require 'stackify_apm/config'
require 'json'
# rubocop:disable Metrics/CyclomaticComplexity
# rubocop:disable Metrics/PerceivedComplexity
module StackifyRubyAPM
  # @api private
  module Spies extend Log

    def self.run_custom_instrumentation
      config = Config.new
      to_instrument = parse_json_config(config.json_config_file)
      if config.logger.nil?
        config.debug_logger
        config.logger.send(:info, '[StackifyRubyAPM] Error: config/stackify.json does not exist which is required for custom instrumentation!') if to_instrument.nil?
      end

      return unless !to_instrument.nil? && !to_instrument.empty? && defined?(to_instrument['instrumentation']) && (to_instrument['instrumentation'].count > 0)

      to_instrument['instrumentation'].each do |custom_spy|
        current_class = defined?(custom_spy['class']) ? custom_spy['class'] : nil
        current_module = defined?(custom_spy['module']) ? custom_spy['module'] : nil
        current_method = custom_spy['method']
        tracked_func = custom_spy['trackedFunction']
        tracked_func_name = defined?(custom_spy['trackedFunctionName']) ? custom_spy['trackedFunctionName'] : ''
        transaction = defined?(custom_spy['transaction']) ? custom_spy['transaction'] : nil
        file_path = defined?(custom_spy['file_path']) ? custom_spy['file_path'] : nil

        if current_class
          tracked_function_tpl = tracked_func_name.nil? ? '{{ClassName}}.{{MethodName}}' : tracked_func_name
          tracked_function_name = tracked_function_tpl.to_s.sub '{{ClassName}}', current_class
          tracked_function_name = tracked_function_name.to_s.sub '{{MethodName}}', current_method
        elsif current_module
          tracked_function_tpl = tracked_func_name.nil? ? '{{ModuleName}}.{{MethodName}}' : tracked_func_name
          tracked_function_name = tracked_function_tpl.to_s.sub '{{ModuleName}}', current_module
          tracked_function_name = tracked_function_name.to_s.sub '{{MethodName}}', current_method
        end

        begin
          if current_module
            if file_path.nil?
              config.logger.send(:info, "[StackifyRubyAPM] Error: Missing file_path in module which is required in custom instrumentation.")
            else
              require file_path
              StackifyRubyAPM::InstrumenterHelper.patched_module(tracked_func, current_module, file_path,
                current_method: current_method, tracked_function_name: tracked_function_name, is_transaction: transaction)
            end
          end
        rescue => e
          throw e
        end

        if !class_exists?(current_class) && !file_path.nil?
          begin
            require file_path
          rescue LoadError => e
            debug "[StackifyRubyAPM] File path doesn't exist."
            debug e.inspect
          end
        end

        # rubocop:disable Style/Next
        if class_exists?(current_class)
          mod_constant = Module.const_get(current_class.to_s)
          klass_method_flag = mod_constant.method_defined?(current_method.to_s)
          singleton_method_flag = mod_constant.respond_to?(current_method.to_s)
          klass_private_method_flag = mod_constant.private_method_defined?(current_method.to_s)
          klass_protected_method_flag = mod_constant.protected_method_defined?(current_method.to_s)

          class_location = mod_constant.instance_methods(false).map do |m|
            mod_constant.instance_method(m).source_location.first
          end.uniq

          class_path = class_location.last

          if klass_method_flag || klass_private_method_flag
            StackifyRubyAPM::InstrumenterHelper.m_class(
              tracked_func,
              current_class,
              class_path,
              current_method: current_method,
              tracked_function_name: tracked_function_name,
              is_transaction: transaction,
              is_private_method: klass_private_method_flag,
              is_protected_method: klass_protected_method_flag
            )
          elsif singleton_method_flag
            StackifyRubyAPM::InstrumenterHelper.m_singleton(tracked_func, current_class, class_path, current_method: current_method, tracked_function_name: tracked_function_name, is_transaction: transaction)
          end
        end
        # rubocop:enable Style/Next
      end
    end
  end
end
# rubocop:enable Metrics/CyclomaticComplexity
# rubocop:enable Metrics/PerceivedComplexity
