# frozen_string_literal: true

#
# Monkey patch for the Logger class for logging log messages.
#
module StackifyRubyAPM
  module Spies
    class LoggerSpy
      NAME = 'logger'
      TYPE = 'core.devlog'

      def install
        Logger.class_eval do
          alias_method 'debug_without_apm', 'debug'
          alias_method 'info_without_apm', 'info'
          alias_method 'warn_without_apm', 'warn'
          alias_method 'error_without_apm', 'error'
          alias_method 'fatal_without_apm', 'fatal'
          alias_method 'unknown_without_apm', 'unknown'

          def debug(progname = nil, &block)
            return debug_without_apm(progname, &block) unless StackifyRubyAPM.current_transaction
            ctx = StackifyRubyAPM::Spies::LoggerSpy.get_log_context('DEBUG', nil, progname, &block)
            StackifyRubyAPM.span NAME, TYPE, context: ctx do
              debug_without_apm(progname, &block)
            end
          end

          def info(progname = nil, &block)
            return info_without_apm(progname, &block) unless StackifyRubyAPM.current_transaction
            ctx = StackifyRubyAPM::Spies::LoggerSpy.get_log_context('INFO', nil, progname, &block)
            StackifyRubyAPM.span NAME, TYPE, context: ctx do
              info_without_apm(progname, &block)
            end
          end

          def warn(progname = nil, &block)
            return warn_without_apm(progname, &block) unless StackifyRubyAPM.current_transaction
            ctx = StackifyRubyAPM::Spies::LoggerSpy.get_log_context('WARN', nil, progname, &block)
            StackifyRubyAPM.span NAME, TYPE, context: ctx do
              warn_without_apm(progname, &block)
            end
          end

          def error(progname = nil, &block)
            return error_without_apm(progname, &block) unless StackifyRubyAPM.current_transaction
            ctx = StackifyRubyAPM::Spies::LoggerSpy.get_log_context('ERROR', nil, progname, &block)
            StackifyRubyAPM.span NAME, TYPE, context: ctx do
              error_without_apm(progname, &block)
            end
          end

          def fatal(progname = nil, &block)
            return fatal_without_apm(progname, &block) unless StackifyRubyAPM.current_transaction
            ctx = StackifyRubyAPM::Spies::LoggerSpy.get_log_context('FATAL', nil, progname, &block)
            StackifyRubyAPM.span NAME, TYPE, context: ctx do
              fatal_without_apm(progname, &block)
            end
          end

          def unknown(progname = nil, &block)
            return unknown_without_apm(progname, &block) unless StackifyRubyAPM.current_transaction
            ctx = StackifyRubyAPM::Spies::LoggerSpy.get_log_context('UNKNOWN', nil, progname, &block)
            StackifyRubyAPM.span NAME, TYPE, context: ctx do
              unknown_without_apm(progname, &block)
            end
          end
        end
      end

      # create and return logger span context
      def self.get_log_context(severity, message, progname)
        ctx = nil

        begin
          log_message = ''
          exception = nil

          if message.nil?
            msg = progname
          else
            msg = message
          end

          case msg
          when ::String
            log_message = msg
          when ::Exception
            log_message = msg.message
            exception = "(#{ msg.class })\n#{ msg.backtrace.join("\n") if msg.backtrace }"
          else
            log_message = msg.inspect
          end

          ctx = Span::Context.new(
            CATEGORY: 'Log',
            SUBCATEGORY: 'Logger',
            LEVEL: severity || 'ANY',
            MESSAGE: log_message,
            PREFIX: 'TRUE'
          )

          if exception
            ctx.EXCEPTION = exception
          end
        rescue Exception => e
          StackifyRubyAPM.agent.error "[LoggerSpy] Error: creating span context."
          StackifyRubyAPM.agent.error "[LoggerSpy] #{e.inspect}"
        end

        ctx
      end
    end

    register 'Logger', 'logger', LoggerSpy.new
  end
end
