# frozen_string_literal: true

# Monkey patch for the Mongo::Monitoring::Global module in mongoDB(http://api.mongodb.com/ruby/current/Mongo/Monitoring/Global.html)
# -> Allows subscribing to events for all Mongo clients.
#

module StackifyRubyAPM
  # @api private
  module Spies
    # @api private
    class MongoSpy
      def install
        ::Mongo::Monitoring::Global.subscribe(
          ::Mongo::Monitoring::COMMAND,
          Subscriber.new
        )
      end

      # @api private
      class Subscriber
        TYPE = 'db.mongodb.query'.freeze

        def initialize
          @events = {}
        end

        def started(event)
          push_event(event)
        end

        def failed(event)
          pop_event(event)
        end

        def succeeded(event)
          pop_event(event)
        end

        private

        def push_event(event)
          return unless StackifyRubyAPM.current_transaction

          # Fetches collection name from Mongo event
          #
          document = event.command.find
          col = nil
          if document
            document.each_with_index do |val, idx|
              col = val[1] if idx.zero?
            end
          end

          # Builds span context
          #
          ctx = Span::Context.new(
            CATEGORY: 'MongoDB',
            MONGODB_COLLECTION: col,
            OPERATION: event.command_name.to_s,
            URL: event.address.to_s
          )

          # Creates new span from Mongo event
          #
          span = StackifyRubyAPM.span(event.command_name, TYPE, context: ctx)
          @events[event.operation_id] = span
        end

        def pop_event(event)
          return unless StackifyRubyAPM.current_transaction

          span = @events.delete(event.operation_id)
          span && span.done
        end
      end
    end

    # Registers Mongo spy, go to: /stackify_apm/spies.rb
    #
    register 'Mongo', 'mongo', MongoSpy.new
  end
end
