# frozen_string_literal: true

# Spies for active record when sinatra framework is used and active_record is being extended. (mysql adapter)

# Not used
# Possible use of ActiveSupport.on_load(:active_record)

require 'stackify_apm/helper/database_helper'

module StackifyRubyAPM
  # @api private
  module Spies
    # @api private
    class MysqlAdapterSpy
      TYPE = 'db.sinatra_active_record.sql'.freeze
      DEFAULT_PORT = 3306
      if ActiveRecord::VERSION::MAJOR.to_i >= 5
        def install
          ActiveRecord::ConnectionAdapters::MySQL::DatabaseStatements.class_eval do
            alias_method 'exec_query_without_apm', 'exec_query'
            alias_method 'exec_delete_without_apm', 'exec_delete'
            alias_method 'exec_update_without_apm', 'exec_update'

            def exec_update(sql, name = nil, binds = [])
              result = nil

              unless StackifyRubyAPM.current_transaction
                exec_update_without_apm(sql, name, binds)
              end

              payload = {sql: sql, binds: binds}
              statement = query_variables(payload)
              check_prepared_stmt(statement, payload)
              ctx = Span::Context.new(statement)

              result = exec_update_without_apm(sql, name, binds)

              StackifyRubyAPM.span name, TYPE, context: ctx do
                return result
              end
            end

            def exec_delete(sql, name = nil, binds = [])
              result = nil

              unless StackifyRubyAPM.current_transaction
                exec_delete_without_apm(sql, name, binds)
              end

              payload = {sql: sql, binds: binds}
              statement = query_variables(payload)
              check_prepared_stmt(statement, payload)
              ctx = Span::Context.new(statement)

              result = exec_delete_without_apm(sql, name, binds)

              StackifyRubyAPM.span name, TYPE, context: ctx do
                return result
              end
            end

            # rubocop:disable Lint/UnusedMethodArgument
            def exec_query(sql, name = 'SQL', binds = [], prepare: false)
              result = nil

              unless StackifyRubyAPM.current_transaction
                exec_query_without_apm(sql, name, binds)
              end

              payload = {sql: sql, binds: binds}
              statement = query_variables(payload)
              check_prepared_stmt(statement, payload)
              ctx = Span::Context.new(statement)

              result = exec_query_without_apm(sql, name, binds)
              StackifyRubyAPM.span name, TYPE, context: ctx do
                return result
              end
            end
            # rubocop:enable Lint/UnusedMethodArgument

            def query_variables(payload)
              props = get_common_db_properties
              props[:PROVIDER] = 'mysql'
              props[:SQL] = payload[:sql]
              props[:URL] = get_host unless !get_host
              props
            end

            def check_prepared_stmt(statement, payload)
              if StackifyRubyAPM.agent.config.prefix_enabled
                check_prepared_stmt_by_placeholder(payload[:sql].include?('?'), statement, payload)
              end
            end

            def get_host
              query_options = self.raw_connection.query_options.to_h
              "#{query_options[:host]}:#{query_options[:port] || DEFAULT_PORT}"
            rescue StandardError => error
              nil
            end
          end
        end
      else
        def install
          ActiveRecord::ConnectionAdapters::Mysql2Adapter.class_eval do
            alias_method 'exec_query_without_apm', 'exec_query'
            alias_method 'exec_delete_without_apm', 'exec_delete'
            alias_method 'exec_update_without_apm', 'exec_update'
            alias_method 'exec_insert_without_apm', 'exec_insert'

            def exec_query(sql, name = 'SQL', binds = [], _prepare: false)
              result = nil

              unless StackifyRubyAPM.current_transaction
                exec_query_without_apm(sql, name, binds)
              end

              ctx = Span::Context.new(
                CATEGORY: 'Database',
                SUBCATEGORY: 'Execute',
                COMPONENT_CATEGORY: 'DB Query',
                COMPONENT_DETAIL: 'Execute SQL Query',
                SQL: sql,
                PROVIDER: 'mysql'
              )
              ctx[:URL] = get_host unless !get_host

              result = exec_query_without_apm(sql, name, binds)

              StackifyRubyAPM.span name, TYPE, context: ctx do
                return result
              end
            end

            def exec_delete(sql, name = nil, binds = [])
              result = nil

              unless StackifyRubyAPM.current_transaction
                exec_delete_without_apm(sql, name, binds)
              end

              ctx = Span::Context.new(
                CATEGORY: 'Database',
                SUBCATEGORY: 'Execute',
                COMPONENT_CATEGORY: 'DB Query',
                COMPONENT_DETAIL: 'Execute SQL Query',
                SQL: sql,
                PROVIDER: 'mysql'
              )
              ctx[:URL] = get_host unless !get_host

              result = exec_delete_without_apm(sql, name, binds)

              StackifyRubyAPM.span name, TYPE, context: ctx do
                return result
              end
            end

            def exec_update(sql, name, binds)
              result = nil

              unless StackifyRubyAPM.current_transaction
                exec_update_without_apm(sql, name, binds)
              end

              ctx = Span::Context.new(
                CATEGORY: 'Database',
                SUBCATEGORY: 'Execute',
                COMPONENT_CATEGORY: 'DB Query',
                COMPONENT_DETAIL: 'Execute SQL Query',
                SQL: sql,
                PROVIDER: 'mysql'
              )
              ctx[:URL] = get_host unless !get_host

              result = exec_update_without_apm(sql, name, binds)

              StackifyRubyAPM.span name, TYPE, context: ctx do
                return result
              end
            end

            def exec_insert(sql, name, binds, _pk = nil, _sequence_name = nil)
              result = nil

              unless StackifyRubyAPM.current_transaction
                exec_insert_without_apm(sql, name, binds, _pk = nil, _sequence_name = nil)
              end

              ctx = Span::Context.new(
                CATEGORY: 'Database',
                SUBCATEGORY: 'Execute',
                COMPONENT_CATEGORY: 'DB Query',
                COMPONENT_DETAIL: 'Execute SQL Query',
                SQL: sql,
                PROVIDER: 'mysql'
              )
              ctx[:URL] = get_host unless !get_host

              result = exec_insert_without_apm(sql, name, binds, _pk = nil, _sequence_name = nil)

              StackifyRubyAPM.span name, TYPE, context: ctx do
                return result
              end
            end

            def get_host
              config = self.config.to_h
              "#{config[:host]}:#{config[:port] || DEFAULT_PORT}"
            rescue StandardError => error
              nil
            end

          end
        end
      end
    end

    if ActiveRecord::VERSION::MAJOR.to_i >= 5
      register 'ActiveRecord::ConnectionAdapters::MySQL::DatabaseStatements', 'active_record/connection_adapters/mysql/database_statements', MysqlAdapterSpy.new
    else
      register 'ActiveRecord::ConnectionAdapters::Mysql2Adapter', 'active_record/connection_adapters/mysql2_adapter', MysqlAdapterSpy.new
    end
  end
end
