# frozen_string_literal: true

# Monkey patch for the Yell::Logger class for logging log messages.
#
module StackifyRubyAPM
  # @api private
  module Spies
    # @api private
    class YellSpy
      SEVERITIES = ['DEBUG', 'INFO', 'WARN', 'ERROR', 'FATAL', 'UNKNOWN'].freeze

      def install
        Yell::Logger.module_eval do
          alias_method 'add_without_apm', 'add'

          # Log messages
          def add(options, *messages, &block)
            return add_without_apm(options, *messages, &block) unless StackifyRubyAPM.current_transaction

            begin
              name = 'yell;'
              type = 'ext.yell'
              log_message = []
              exception = []

              messages.each { |msg|
                case msg
                when ::String
                  log_message.push(msg)
                when ::Exception
                  log_message.push(msg.message)
                  exception.push("(#{ msg.class })\n#{ msg.backtrace.join("\n") if msg.backtrace }")
                else
                  log_message.push(msg.inspect)
                end
              }

              ctx = Span::Context.new(
                CATEGORY: 'Log',
                SUBCATEGORY: 'Yell',
                LEVEL: SEVERITIES[options.severity],
                MESSAGE: log_message.to_s,
                PREFIX: 'TRUE'
              )

              if exception.length > 0
                ctx.EXCEPTION = exception.to_s
              end
            rescue Exception => e
              StackifyRubyAPM.agent.error "[YellSpy] Error: creating span context."
              StackifyRubyAPM.agent.error "[YellSpy] #{e.inspect}"
              return add_without_apm(options, *messages, &block)
            end

            StackifyRubyAPM.span name, type, context: ctx do
              add_without_apm(options, *messages, &block)
            end
          end
        end
      end
    end

    register 'Yell', 'yell', YellSpy.new
  end
end
