# frozen_string_literal: true

# The Subscriber registers the existing events.
# This will trigger (together with the Middleware) once there is a web request.

require 'active_support/notifications'
require 'stackify_apm/normalizers'

module StackifyRubyAPM
  # @api private
  class Subscriber
    include Log

    def initialize(agent)
      debug '[Subscriber] initialize(agent)' if ENV['STACKIFY_TRANSPORT_LOG_LEVEL'] == '0'
      @agent = agent
      @normalizers = Normalizers.build(agent.config)
    end

    def register!
      unregister! if @subscription

      @subscription =
        ActiveSupport::Notifications.subscribe(notifications_regex, self)
    end

    def unregister!
      ActiveSupport::Notifications.unsubscribe @subscription
      @subscription = nil
    end

    # AS::Notifications API

    Notification = Struct.new(:id, :span)

    # [call] Called when the rails version is 3.x
    # rubocop:disable Metrics/CyclomaticComplexity
    # rubocop:disable Metrics/PerceivedComplexity
    def call(name, started, finished, id, payload)
      return unless (transaction = @agent.current_transaction)

      normalized = @normalizers.normalize(transaction, name, payload)

      if started
        span =
          if normalized == :skip
            nil
          else
            name, type, context = normalized
            @agent.span(name, type, context: context)
          end
        transaction.notifications << Notification.new(id, span)
      end
      # rubocop:enable Metrics/CyclomaticComplexity
      # rubocop:enable Metrics/PerceivedComplexity
      # rubocop:disable Style/GuardClause

      if finished
        while (notification = transaction.notifications.pop)
          next unless notification.id == id
          if (span = notification.span)
            span.done
          end
          return
        end
      end
      # rubocop:enable Style/GuardClause
    end

    # [start] Called when the rails version is NOT 3.x
    def start(name, id, payload)
      return unless (transaction = @agent.current_transaction)

      normalized = @normalizers.normalize(transaction, name, payload)

      span =
        if normalized == :skip
          nil
        else
          name, type, context = normalized
          @agent.span(name, type, context: context)
        end

      transaction.notifications << Notification.new(id, span)
    end

    # [finish] Called when the rails version is NOT 3.x
    def finish(_name, id, _payload)
      return unless (transaction = @agent.current_transaction)

      while (notification = transaction.notifications.pop)
        next unless notification.id == id

        if (span = notification.span)
          span.done
        end
        return
      end
    end

    private

    def notifications_regex
      @notifications_regex ||= /(#{@normalizers.keys.join('|')})/
    end
  end
end
