# frozen_string_literal: true

#
# This class creates and initializes a new transaction.
#
module StackifyRubyAPM
  require 'securerandom'
  # @api private
  class Transaction
    DEFAULT_TYPE = 'custom'.freeze

    def initialize(instrumenter, name = nil,
                   type = nil, context: nil)
      # puts "Loads transaction new initialize
      # instrumenter, name = nil, type = nil, context: nil"
      @id = SecureRandom.uuid
      @instrumenter = instrumenter
      @name = name
      @type = type || DEFAULT_TYPE
      @timestamp = Time.now.to_f * 1000

      @spans = []
      @span_id_ticker = -1
      @dropped_spans = 0

      @notifications = [] # for AS::Notifications
      @context = context || Context.new
      @exceptions = []

      yield self if block_given?
    end

    attr_accessor :name, :type, :http_status
    attr_reader :id, :context, :duration, :dropped_spans, :instrumenter
    attr_reader :timestamp, :spans, :result, :notifications, :exceptions

    def add_exception(exception)
      @exceptions << exception
    end

    def release
      @instrumenter.current_transaction = nil
    end

    def done(result = nil)
      @duration = Time.now.to_f * 1000
      @result = result
      @http_status = result

      self
    end

    def done?
      !@duration.nil?
    end

    def submit(result = nil, status: nil, headers: {})
      done result unless duration
      context.response = Context::Response.new(status, headers: headers) if status

      release
      @instrumenter.submit_transaction self

      self
    end

    # This method is being used in unit testing
    def running_spans
      spans.select(&:running?)
    end

    def span(name, type = nil, backtrace: nil, context: nil)
      span = build_and_start_span(name, type, context, backtrace)
      return span unless block_given?

      begin
        result = yield span
      ensure
        span.done
      end

      result
    end

    def current_span
      spans.reverse.lazy.find(&:running?)
    end

    def inspect
      "<StackifyRubyAPM::Transaction id:#{id}" \
      " name:#{name.inspect}" \
      " type:#{type.inspect}" \
      '>'
    end

    private

    def next_span_id
      @span_id_ticker += 1
    end

    def next_span(name, type, context)
      Span.new(
        self,
        next_span_id,
        name,
        type,
        parent_id: current_span.nil? ? -1 : current_span.id,
        context: context,
        http_status: @http_status
      )
    end

    def span_frames_min_duration?
      @instrumenter.agent.config.span_frames_min_duration != 0
    end

    def build_and_start_span(name, type, context, backtrace)
      span = Span.new(
        self,
        next_span_id,
        name,
        type,
        parent_id: current_span.nil? ? -1 : current_span.id,
        context: context,
        http_status: @http_status
      )

      # span = next_span(name, type, context)
      spans << span

      span.original_backtrace = backtrace if backtrace && span_frames_min_duration?

      span.start
    end
  end
end
