# frozen_string_literal: true

require 'uri'
require 'faraday'
require 'ostruct'

module StackifyRubyAPM
  # This class will handle the sending of transaction messages through HTTP.
  # @api private
  class AgentHTTPClient < AgentBaseTransport
    include Log

    def initialize(config)
      @config = config
      super(config)
    end

    def get_protobuf_headers
      {
        'Content-Type' => 'application/x-protobuf'
      }.freeze
    end

    def get_json_headers
      {
        'Content-Type' => 'application/json'
      }.freeze
    end

    # rubocop:disable Metrics/CyclomaticComplexity
    # rubocop:disable Metrics/PerceivedComplexity
    #
    # This method will send a transction message to HTTP request.
    # It will accept Array of transactions.
    def post(transactions = [])
      debug '[AgentHTTPClient] post()' if ENV['STACKIFY_TRANSPORT_LOG_LEVEL'] == '0'
      return unless ENV['STACKIFY_RUBY_ENV'] != 'rspec'
      max_retries = @config.max_retries
      retry_count = 0
      delay = @config.delay_seconds
      begin
        message = get_json_message(transactions)
        conn = Faraday.new(ssl: { verify: false })
        response = conn.post do |req|
          req.url URI(@config.transport_http_endpoint + @config.agent_traces_url).to_s
          req.headers = get_json_headers
          req.body = message
        end
        if defined?(response.status) && response.status == 200
          debug '[AgentHTTPClient] Successfully send message via http request.' if ENV['STACKIFY_TRANSPORT_LOG_LEVEL'] == '0'
        elsif ENV['STACKIFY_TRANSPORT_LOG_LEVEL'] == '0'
          debug "[AgentHTTPClient] Failure sending via http request: #{response.inspect}"
        end
      rescue StandardError => e
        debug '[AgentHTTPClient] All retries are exhausted!' if retry_count >= max_retries
        retry_count += 1
        if retry_count < max_retries
          debug "[AgentHTTPClient] post() exception: #{e.inspect}"
          debug "[AgentHTTPClient] An error occured. Retries left: #{max_retries - retry_count}"
        end
        sleep delay += retry_count
        retry if retry_count < max_retries + 1
      end
    end
    # rubocop:enable Metrics/CyclomaticComplexity
    # rubocop:enable Metrics/PerceivedComplexity
  end
end
