# frozen_string_literal: true

# This class will receive the transactions and send it to HTTP/TraceLogger.
# Responsible for sending transactions to APM for every 1 secs (config.flush_interval_seconds default)
# Worker will constantly execute transaction sending to APM until the Agent is stopped.

require 'concurrent/timer_task'

module StackifyRubyAPM
  # @api private
  class Worker
    include Log

    # @api private
    class StopMsg; end

    # @api private
    class FlushMsg; end

    # @api private
    class ErrorMsg
      def initialize(error)
        @error = error
      end

      attr_reader :error
    end

    def initialize(config, messages, pending_transactions, trace_logger)
      @config = config
      @messages = messages
      @pending_transactions = pending_transactions
      @trace_logger = trace_logger

      @serializers = Struct.new(:transactions, :errors).new(
        Serializers::Transactions.new(config),
        Serializers::Errors.new(config)
      )
    end

    attr_reader :config, :messages, :pending_transactions

    # Collects and sends the transactions
    #
    def run_forever
      @timer_task = build_timer_task.execute

      while (msg = messages.pop)
        case msg
        when ErrorMsg
          post_error msg
        when FlushMsg
          collect_and_send_transactions
        when StopMsg
          # empty collected transactions before exiting
          collect_and_send_transactions
          stop!
        end
      end
    end

    private

    def stop!
      @timer_task && @timer_task.shutdown
      Thread.exit
    end

    # flush_interval_seconds - interval with which transactions should be sent to the APM.
    # The running task responsible for initiating the transactions sending
    #
    def build_timer_task
      Concurrent::TimerTask.new(execution_interval: config.flush_interval_seconds) do
        messages.push(FlushMsg.new)
      end
    end

    # rubocop:disable Lint/RescueException
    # Collects, builds, and sends transactions via HTTP/TraceLogger
    #
    def collect_and_send_transactions
      return if pending_transactions.empty?
      transactions = collect_batched_transactions
      debug '[Worker] collect_and_send_transactions() transactions : ' + transactions.inspect.to_s if ENV['STACKIFY_TRANSPORT_LOG_LEVEL'] == '0'
      begin
        @trace_logger.post(transactions)
      rescue ::Exception => e
        error '[Worker] Collect_and_send_transactions error:'
        error e.inspect
        error e.backtrace.join("\n")
        nil
      end
    end
    # rubocop:enable Lint/RescueException

    def post_error(_msg)
      return if pending_transactions.empty?
      # transactions = collect_batched_transactions
      # payload = @serializers.errors.build_all([msg.error])
      # @trace_logger.post(payload, transactions[0])
    end

    # rubocop:disable Lint/HandleExceptions
    def collect_batched_transactions
      batch = []

      begin
        while (transaction = pending_transactions.pop(true)) &&
              batch.length <= config.max_queue_size
          batch << transaction
        end
      rescue ThreadError # queue empty
      end

      batch
    end
    # rubocop:enable Lint/HandleExceptions
  end
end
