# frozen_string_literal: true

#
# The entry point once the server starts after the Stackify APM has been integrated.
#
# Checks if Web App's framework is Rails or Sinatra:
#   if Rails
#     -> call lib/stackify_apm/railtie.rb to start Agent
#   elsif Sinatra
#     -> access the Web App's /config.ru file to start Agent
#   end
#
# Incoming requests are then handled by the Middleware (middleware.rb)
#

require 'stackify_apm/log'
require 'stackify_apm/version'
require 'stackify_apm/util/dig'
# Checks ruby version
require RUBY_VERSION.to_f >= 2.4 ? 'stackify_apm/logger/logger_high_version' : 'stackify_apm/logger/logger_lower_version'
require 'stackify_apm/agent'
require 'stackify_apm/helper/database_helper'
require 'stackify_apm/util'
require 'stackify_apm/config'
require 'stackify_apm/context'
require 'stackify_apm/instrumenter'
require 'stackify_apm/internal_error'
require 'stackify_apm/response_manipulator'
require 'stackify_apm/middleware'
require 'stackify_apm/instrumenter_helper'
require 'stackify_apm/transport_selector'
require 'stackify_apm/transport/agent_base'
require 'stackify_apm/transport/log_client'
require 'stackify_apm/transport/unix_socket_client'
require 'stackify_apm/transport/agent_http_client'
require 'stackify_apm/transport/aws_lambda_logging'

# Checks if the framework using is Rails
require 'stackify_apm/railtie' if defined?(::Rails::Railtie)

# Start medoule for StackifyRubyAPM Agent
module StackifyRubyAPM
  include Log
  # Starts the StackifyRubyAPM Agent
  #
  # @param config [Config] An instance of Config
  # @return [Agent] The resulting [Agent]
  def self.start(config = {})
    Agent.start config
  end

  def self.inject_rum_script
    config = StackifyRubyAPM::Config.new
    client_id = config.client_id
    device_id = config.device_id
    client_rundomain = config.client_run_domain
    transaction_id = defined?(StackifyRubyAPM.current_transaction.id) ? StackifyRubyAPM.current_transaction.id : nil
    inject_flag = false

    if client_id && device_id
      inject_flag = true if check_isdomain(client_rundomain)
    end

    return unless inject_flag
    data_request_id = "V2|#{transaction_id}|#{client_id}|#{device_id}"
    @rum_script_injected = true
    "<script src=\"#{config.rum_script_src}\" data-host=\"#{client_rundomain}\" data-requestId=\"#{data_request_id}\"
    data-a=\"#{config.application_name}\" data-e=\"#{config.environment_name}\" data-enableInternalLogging=\"#{config.debug_logging}\" type=\"text/javascript\" async></script>"
  end

  # Run the custom instrument
  def self.run_custom_instrument
    return unless running?
    StackifyRubyAPM::InstrumenterHelper.custom_instrumented_reset
    StackifyRubyAPM::Spies.run_custom_instrumentation
  end

  # Stops the StackifyRubyAPM Agent
  def self.stop
    Agent.stop
  end

  # @return [Boolean] Whether there's an [Agent] running
  def self.running?
    Agent.running?
  end

  # @return [Agent] Currently running [Agent] if any
  def self.agent
    Agent.instance
  end
  ### Metrics

  # Returns the currently active transaction (if any)
  #
  # @return [Transaction] if any
  def self.current_transaction
    agent && agent.current_transaction
  end

  # Start a new transaction or return the currently running
  #
  # @param name [String] A description of the transaction, eg
  # `ExamplesController#index`
  # @param type [String] The kind of the transaction, eg `app.request.get` or
  # `db.mysql2.query`
  # @param context [Context] An optional [Context]
  # @yield [Transaction] Optional block encapsulating transaction
  # @return [Transaction] Unless block given
  def self.transaction(name = nil, type = nil, context: nil, &block)
    return (block_given? ? yield : nil) unless agent

    agent.transaction(name, type, context: context, &block)
  end

  # Starts a new span under the current transaction
  #
  # @param name [String] A description of the span, eq `SELECT FROM "users"`
  # @param type [String] The kind of span, eq `db.mysql2.query`
  # @param context [Span::Context] Context information about the span
  # @yield [Span] Optional block encapsulating span
  # @return [Span] Unless block given
  def self.span(name, type = nil, context: nil, include_stacktrace: true,
                &block)
    return (block_given? ? yield : nil) unless agent

    agent.span(
      name,
      type,
      context: context,
      backtrace: include_stacktrace ? caller : nil,
      &block
    )
  end

  # Build a [Context] from a Rack `env`. The context may include information
  # about the request, response, current user and more
  #
  # @param rack_env [Rack::Env] A Rack env
  # @return [Context] The built context
  def self.build_context(rack_env)
    agent && agent.build_context(rack_env)
  end

  ### Errors

  # Report and exception to APM
  #
  # @param exception [Exception] The exception
  # @param handled [Boolean] Whether the exception was rescued
  # @return [Error] The generated [Error]
  def self.report(exception, handled: true)
    agent && agent.report(exception, handled: handled)
  end

  # Report a custom string error message to APM
  #
  # @param message [String] The message
  # @return [Error] The generated [Error]
  def self.report_message(message, **attrs)
    agent && agent.report_message(message, backtrace: caller, **attrs)
  end

  # Check if the URL is a domain
  #
  # @param message [String] The message
  # @return TRUE/FALASE
  def self.check_isdomain(url)
    url =~ %r{^(http|https):\/\/|[a-z0-9]+([\-\.]{1}[a-z0-9]+)*\.[a-z]{2,6}(:[0-9]{1,5})?(\/.*)?$}
  end

  # Check if the RUM Script is Injected
  #
  # @return TRUE/FALASE
  def self.rum_script_inject
    @rum_script_injected
  end
end
